#!/usr/local/bin/python
#
# a python class to interface with cpwright's cmserver
#    I don't know what protocol version this starts working at.. i'm using 0.3.3 to dev here, so that should work
#
# 0.8.1, Fri Mar 19 15:24:06 1999
# johnchia@innocent.com
#

from string import split,join,find,lower,rfind
from time import sleep
from telnetlib import *
from socket import error

class pppctlClass:

	"PPP Control Class for cpwright's c-mserver."

	def __init__(self, host=0, port=0, passwd=0, user=0):
		self.my_buf = ""
		self.err_msg = ""
		self.in_buf = ""

		self.timeout = 60		# default timeout for recv'ing (in seconds)
						# ** Note this has to be extremely high for mserver; should be maybe double
						# how long your modem takes to dial..

		self.pollrate = 0.03		# how many seconds to wait between each poll
						# used when self.return_when_connected is true

		self.return_when_complete = 1	# return only when completely connected, or
						# completely down, set to 0 for returning right 
						# after the command is sent (start_<_0.8-esque)

		self.lprompt = ""		# lower-cased prompt

		if host and port and passwd:
			self.open_ppp( host, port, passwd, user )


	def open_ppp ( self, host, port, passwd = 0, user = 0, prompt = "READY\012"):
		""" 	open_ppp - Connect and authenticate
			returns:
				-1 error connecting
			 	0 bad password, closed
			 	1 open, authenticated, ready.	"""
		try:
			self.ppp = Telnet(host, port)		# open host
		except self.socket.error:
			return(-1)				# exit if error

		self.my_buf = self.ppp.read_until(prompt, self.timeout)	# read until prompt

		if passwd and user:
			self.ppp.write( 'USER:' + user + '\n' )			# send username
			self.my_buf = self.ppp.read_until(prompt, self.timeout)	# read until prompt

			self.ppp.write( 'PASS:' + passwd + '\n' )		# send password
			self.my_buf = self.ppp.read_until(prompt, self.timeout)	# read until prompt

			self.ppp.write('AUTH\n')				# send password
			self.my_buf = self.ppp.read_until(prompt, self.timeout)	# read until prompt

			if find( self.my_buf, "NOT AUTH" ) > -1: return(0)	# make sure password went through OK

		return(1)


	# --
	# send_ppp - send data out the pipe
	# returns what was sent back minus prompt

	def send_ppp ( self, data, prompt = "READY\012" ):
		self.in_buf = self.ppp.read_very_eager()

		self.ppp.write( data + "\n" )			# Only mserver needs the \n, odd?

		self.in_buf = ['']
		while self.in_buf == ['']:
			self.in_buf = self.ppp.read_until(prompt,self.timeout)

		if find(self.in_buf,"BEGIN") > -1:
			self.in_buf = self.in_buf[ rfind( self.in_buf, "BEGIN\n" ) + len("BEGIN\n") : find( self.in_buf, "END" )-1 ]
		else:
			self.in_buf = self.in_buf[ : find( self.in_buf, "\n" ) ]

		return(self.in_buf)


	# --
	# Throughput Function - get one link's throughput (badly named)
	# returns 3,4,5,6 down there

	def throughput(self):

	        my_buf = self.send_ppp ( "NETLOAD" )
		my_buf = split(my_buf,'\n')	# split by \n

	        bytein	= split( my_buf[0] ) [1]	# bytes in
	        byteout	= split( my_buf[1] ) [1]	# bytes out


		link_up	= int( self.send_ppp("CTIME") ) - int( self.send_ppp("TIME") )	# link uptime in seconds (I think)


	        peak_on	= 0	# date of peak


		return( bytein, byteout, peak_on, int(link_up) )


	# --
	# status - get all links and their status
	# For mserver:
	# 	0 <name> 
	#	1 <open/dial/closed/etc)
	#	2 [connect speed]
	#	3 [bytes in] 
	#	4 [bytes out] 
	#	5 [date of peak] (will be 0)
	#	6 [connect time]
	#	-- indexes 2-6 are returned only when connection is open.

	def status(self):

		out = []
		self.my_buf = split( self.send_ppp('LIST'), "\n" )
		link_status = split( self.send_ppp('STAT'), ":" )

		if link_status[0] == "UP":

		        for x in range( len( self.my_buf ) ):
				out.append([])

				out[x].append( self.my_buf[x] )		# put name -> out

				if self.my_buf[x] == link_status[1]:

					out[x].append( "open" )		# put "UP" -> out
					out[x].append( link_status[2] ) # put connect speed -> out

					for a in range(4): out[x].append("")
					out[x][3], out[x][4], out[x][5], out[x][6] = self.throughput()	# Get throughput

				else:
					out[x].append( "closed" )	# put "DOWN" -> out

		else:							# if it's down, append just names and "DOWN"s
			for x in range( len( self.my_buf ) ):
				out.append([])
				out[x].append( self.my_buf[x] )
				out[x].append( "closed" )

		return(out)


	# --
	# get_status - get one link's status (or in the case of mserver, if a link is up or not)
	# returns status (open,dial,closed,etc) of conn

	def get_status(self, conn=0):
		status = split( self.send_ppp('STAT'), ":" )[0]
		if status == "UP": return( 'open' )
		elif status == "DOWN": return( 'closed' )
		else: return( 'err' )

	# --
	# dial_ppp - dial a link
	# returns 1 for success, 0 for failure, error mesg stored in self.err_msg

	def dial_ppp ( self, conn ):
	
		self.my_buf = self.send_ppp ( 'DIAL:' + conn )	# send command

		if find(self.my_buf, "Connected") == -1:	# there was some sort of error.
			self.err_msg = self.my_buf
			return(0)

		if not self.return_when_complete: return(1)	# return right away if that's
								# what we're supposed ta do

		sleep(self.pollrate)				# sleep a second to make sure we're dialing..
		status = "null"
		while status != "open" and status != "err" and status != "closed":
			sleep(self.pollrate)
			status = self.get_status()
		if status == "UP": return(1)
		else: return(0)


	# --
	# down_ppp - down a link
	# Returns 1 for success

	def down_ppp( self, conn = 0):
		self.my_buf = self.send_ppp ( 'KILL' )			# send command

		if self.my_buf != "SUCESS":					# there was some sort of error.
			self.err_msg = self.my_buf
			return(1)

		if not self.return_when_complete: return(0)		# return right away if that's
									# what we're supposed ta do

		sleep(self.pollrate)					# sleep a second to make sure we've begun dialing
		status = "null"
		while status != "open" and status != "err" and status != "closed":
			sleep(self.pollrate)
			status = self.get_status(conn)
		if status == "closed": return(1)
		else: return(0)


	# --
	# close_ppp - closes connection
	# disconnects from the PPP control server

	def close_ppp(self):
		self.ppp.close()


"""
#
# Connection List
# replaced with better status function
# returns a list of connections
#
def conn_list():
	conns = s_r("show links\n")

	conns.reverse()
	conns.remove(conns[0])
	conns.reverse()
	
	for x in range(len(conns)):
		y = conns[x]
		y = self.split(y)
		conns[x] = y[1]

	return(conns)

States (from /usr/src/usr.sbin/ppp/datalink.c freebsd-STABLE 3.1-199902XX)
  "closed",
  "opening",
  "hangup",
  "dial",
  "login",
  "ready",
  "lcp",
  "auth",
  "cbcp",
  "open"

from mstart_ppp import *
link = pppctlClass()
link.open_ppp('eye',222,'','kid')



"""

